/*
 *
 *	    M U L T I - D E V I C E   Q U E U E I N G   S Y S T E M
 *
 *			   Douglas P. Kingston III
 *
 *			Batch Queue Server  (SHSERVER)
 *
 *	Note that a pipe to the Daemon will be on file descriptor 3.
 *	________________________________________________________________
 *
 *				Revision History
 *				================
 *	17 Sep 82  DPK	Added code to support specification of an
 *			output file for FDs 1 and 2.
 *
 *	22 Sep 82  DPK	Modified code that execs shells to allow invoking
 *			shells with arguments.
 *
 *	21 Dec 82  DPK	Increased size of linebuf to handle large environment
 *			variables (e.g. TERMCAP).
 *
 *	10 Aug 89  DAG	Overhauled, to improve portability and fix bugs.
 */

#include <errno.h>
#include <signal.h>
#include <stdio.h>

extern void	_exit(), exit();
extern char	*malloc(), *strcat(), *strcpy();
extern int	atoi(), chdir(), close(), dup(), fork(), execve(), nice(),
		open(), strlen(), unlink(), wait();
extern unsigned	sleep();
extern long	lseek();

#if V6

extern int	kill();
#define	killpg( pid, sig )	kill( pid, sig );

#else

extern int	umask();

#if SYS3 || V7

extern int	kill();
#define	killpg( pid, sig )	kill( 0, sig );

#else	/* BSD42 || BSD41 || BSD41c || BRLUNIX */

extern int	killpg();		/* assumed to be in C library */

#endif	/* SYS3 || V7 */
#endif	/* V6 */

#ifndef _NFILE

extern int	getdtablesize();
#define	_NFILE	getdtablesize()

#endif

#ifndef SIGTERM
#define	SIGTERM	15			/* MDQS usurps this signal */
#endif


#include "queue.h"
#include "qreplys.h"

extern void		rdconf(), retmsg();
extern int		parse();
extern struct request	*gethdr();

extern char		*Qdatadir;


void			doflush(), dorestart();

static void		bug(), fatal(), hold(), restart(), setenv(), term();
static char		*xmalloc(), *xstrdup();


#ifndef NENVVARS
#define NENVVARS	1000	/* 1000 environment variables is plenty */
#endif

static char		*envp[NENVVARS] = { 0 };
static int		nused = 0;

static char		pname[] = "SHSERVER";

static char		*shell = "/bin/sh";	/* shell to be exec()ed */
static char		*user_shell = NULL;	/* user-specified shell */
static char		*output = "/dev/null";
static char		*workdir = "/tmp";

static char		cmdfile[PATHSIZ] = { 0 }; /* MDQS copy of input file */
static int		fin = -1;	/* input FD for above */

static int		Batchnice = 19;	/* default nice() value */
static struct paramtab	paramtab[] =
	{
	"batch-nice",	INTVAR,		(char **)&Batchnice,	NULL,
	0,		0,		0,			0
	};

static int		pid = 0;	/* child (shell) process ID */


/*
 *		M A I N  ( )
 *  As an MDQS queue server, this process will be started up
 *  with the queue control file as the standard input, and the
 *  device (should be /dev/null for shserver) on the standard output.
 */

main( argc, argv )
	int		argc;
	char		*argv[];
{
	char		**av;		/* argv for exec(), set up by parse() */
	int		status;		/* exit status from wait() */
	register int	i;		/* used for a variety of things */

	signal( SIGHUP, SIG_IGN );	/* Paranoid. */
	signal( SIGEMT, dorestart );
	signal( SIGTERM, doflush );

	rdconf( paramtab );		/* may update Batchnice */

	{
	struct request	r;		/* unpacked user request data */

	if ( gethdr( stdin, &r ) == NULL )
		bug( "Error reading request header" );

	/* Request header skipped; otherwise unused. */
	}

	{
	static char	linebuf[4 * LINESIZ];	/* room for TERMCAP etc. */

	while ( fgets( linebuf, (int)sizeof linebuf, stdin ) != NULL )
		{
		register char	*cp;	/* -> linebuf[.] */

		if ( *(cp = &linebuf[strlen( linebuf ) - 1]) == '\n' )
			*cp = '\0';

		cp = &linebuf[0];
		switch ( *cp++ )
			{
		case 'D':		/* working directory */
			workdir = xstrdup( cp );
			break;

		case 'E':		/* environment string */
			setenv( cp );
			break;

		case 'I':		/* spooled file, used for shell input */
			sprintf( cmdfile, "%s/%s", Qdatadir, cp );

			if ( (fin = open( cmdfile, READ )) < 0 )
				bug( "Can't open command file" );

			break;

		case 'P':		/* file to receive shell output */
			output = xstrdup( cp );
			break;

		case 'S':		/* shell to interpret script */
			user_shell = shell = xstrdup( cp );
			break;

		default:		/* ignore unsupported control lines */
			break;
			}
		}
	}

	if ( fin < 0 )			/* 'I' control line is required */
		bug( "Shell input file wasn't specified" );

	if ( parse( shell, &av ) < 0 )
		fatal( "Error in parsing shell name" );

#if !V6
	umask( 077 );			/* Secure. */
#endif

	close( 1 );

	if ( (i = open( output, 1 )) != 1
	  && (i >= 0
	   || errno != ENOENT
	   || creat( output, 0600 ) != 1	/* Secure. */
	     )
	   )
		fatal( "Couldn't open file for shell output" );

	lseek( 1, 0L, 2 );			/* end of the file (append) */

	/*
	 *	Now we are ready to start the shell to interpret the batch
	 *	commands.  We are already running with the same UID/GID
	 *	as the original "batch" or "qpr" invoker.
	 */

	fflush( stderr );

	if ( (pid = fork()) == -1 )
		restart( "Fork failed" );
	else if ( pid == 0 )
		{			/* Child */
		if ( chdir( workdir ) < 0 )
			chdir( "/tmp" );	/* It's a safe default. */

		close( 0 );
		dup( fin );			/* Standard input is cmdfile. */

		close( 2 );
		dup( 1 );

		for ( i = _NFILE - 1; i > 2; --i )
			close( i );

		nice( Batchnice );

		execve( *av, av, envp );	/* use shell to interpret */

		_exit( 0x77 );
		}

	/* Parent (shserver) process continues: */

	while ( (i = wait( &status )) != pid && !(i < 0 && errno == EINTR) )
		;

	if ( i < 0 )
		bug( "Wait for shell failed" );

	if ( (status & 0xFF) != 0 )
		hold( "Shell died" );
	else if ( (status & 0xFF00) != 0 )
		if ( (status & 0xFF00) != 0x7700 )
			fatal( "Nonzero script exit status" );
		else
			if ( user_shell != NULL )
				fatal( "Exec of specified shell failed" );
			else
				hold( "Standard shell exec failed" );
	else
		term( 0, RP_OK, (char *)0 );
	}

/*
 *		S E T E N V  ( )
 *	This function is responsible for building the environment
 *	variable vector.  It is passed a string to be added to the
 *	(initially empty) environment being constructed.
 */

static void
setenv( envstr )
	char	*envstr;
	{
	if ( nused < NENVVARS - 1 )
		envp[nused++] = xstrdup( envstr );
	else
		fatal( "Too many environment vars" );
	}

/*
 *		X M A L L O C  ( )
 *	Like malloc(), but it always returns a valid pointer or exits.
 */

static char *
xmalloc( size )
	int	size;
	{
	register char	*cp;

	if ( (cp = malloc( size )) == NULL )
		hold( "No free memory" );

	return( cp );
	}

/*
 *	D O R E S T A R T  ( )
 *
 *  This function is called by a signal to restart a job.
 */

void
dorestart( sig )
	int	sig;
	{
	signal( sig, SIG_IGN );

	retmsg( RP_RESTART, "RESTART received" );

	if ( pid != 0 )
		{
		signal( SIGTERM, SIG_IGN );
		killpg( pid, SIGTERM );
		sleep( 3 );
		killpg( pid, SIGKILL );
		}

	exit( 0 );
	}

/*
 *	D O F L U S H  ( )
 *
 *  This function is called by a signal to "flush" the job.
 */
void
doflush( sig )
	int	sig;
	{
	signal( sig, SIG_IGN );

	retmsg( RP_FLUSH, "Server received FLUSH request" );

	if ( pid != 0 )
		{
/*		signal( SIGTERM, SIG_IGN );	/* redundant */
		killpg( pid, SIGTERM );
		sleep( 3 );
		killpg( pid, SIGKILL );
		}

	exit( 0 );
	}

/* xstrdup -- duplicate a string; abort (via xmalloc()) if not enough memory */

static char *
xstrdup( str )
	char	*str;
	{
	return strcpy( xmalloc( strlen( str ) + 1 ), str );
	}

/* restart -- reschedule slave for later execution */

static void
restart( msg )
	char	*msg;
	{
	term( 0, RP_RESTART, msg );	/* put back into queue */
	}

/* hold -- log recoverable MDQS error on standard error output, then die */

static void
hold( msg )
	char	*msg;
	{
	fprintf( stderr, "%s: %s\n", pname, msg );
	term( 0, RP_HOLD, msg );	/* hold for MDQS adminstrator */
	}

/* bug -- log fatal MDQS error on standard error output, then die */

static void
bug( msg )
	char	*msg;
	{
	fprintf( stderr, "%s: %s\n", pname, msg );
	fatal( msg );
	}

/* fatal -- die upon fatal user (not MDQS) error */

static void
fatal( msg )
	char	*msg;
	{
	term( 99, RP_FATAL, msg );
	}

/* term -- terminate server */

static void
term( status, code, msg )
	char	*msg;
	{
	if ( cmdfile[0] != '\0' )
		unlink( cmdfile );

	retmsg( code, msg );
	exit( status );
	}
