static char	rcsid[] = "$Source: vplottest.v$$Revision: 1.10$";
/*  VPLOT: version %I%				updated %G%
 *
 *  Reads standard graphics input and produces a plot on the
 *  Varian or Versatec
 *  -- creates in /usr/tmp a raster file of 500 1K-blocks.
 *	Defining QUE_RASTER compiles version that queues raster file
 *	using qpr(1).
 * $Log:	vplottest.v $
 * Revision 1.10  83/05/07  13:34:42  dave
 * resetting of scaling should only be done once per plotfile
 * 
 * Revision 1.9  83/05/07  13:25:12  dave
 * needed to clear done1 flag when starting a new plotfile
 * 
 * Revision 1.8  83/05/07  12:37:49  dave
 * set buffer size for raster dump to 16*512.
 * 
 * Revision 1.7  83/05/07  11:31:35  dave
 * added unlink for raster file before normal exit.
 * 
 * Revision 1.6  83/05/07  09:49:35  dave
 * defined macros for bit/byte/block for a point
 * TEST2 version seems to work now with corrected raster calculations
 * so raster file does not need special processing on output using
 * this version.
 * 
 * Revision 1.5  83/05/02  13:13:23  dave
 * conditional compilation for stand-alone versions & raster-file queuing
 * using qpr(1).
 * 
 */
#include <stdio.h>
#include <signal.h>
#include <vfont.h>
#include <sys/vcmd.h>

#define VAX	1	/* Machine Flag (don't simulate VM on vax) */
#define	TEST2	1
#define	VERSATEC_V80_211	1	/* we have this one */

#define BSIZ	512		/* Size of blocks */
#ifndef QUE_RASTER
#define IO_BSIZ	16*512		/* # of blocks to read/write for rasterdmp */
#endif QUE_RASTER

#ifdef VERSATEC_V80_211
#define BYTES_PER_LINE	264	/* # of bytes per raster line of versatec */
#define RASTERSIZE	BYTES_PER_LINE*8	/* # bits per raster line */
#define TOTALBLOCKS	(BYTES_PER_LINE*RASTERSIZE/BSIZ)+1
#endif

#ifdef VAX
#define NB	TOTALBLOCKS	/* Number of blocks in virtual memory */
#else
	NB	88		/* Number of blocks kept in memory */
#endif
#define	mapx(x)	((RASTERSIZE*((x)-botx)/del)+centx)
#define	mapy(y)	((RASTERSIZE*(del-(y)+boty)/del)-centy)
#ifndef TEST2
#define	pointbyte(x,y) (((y&077)<<3)+((x>>3)&07))
#define	pointbit(x,y) (1<<(7-(x&07)))
#define blockno(x,y)	((x>>6)+((y&~077)>>1))
#else TEST2
#define	pointbyte(x,y) (((y<<3)+(y<<8)+(x>>3))&0777)
#define	pointbit(x,y) (128>>(x&07))
#define blockno(x,y)	(((y<<3)+(y<<8)+(x>>3))>>9)
/* #define blockno(x,y)	((y>>1)+(y>>6)+(x>>12)) */
#endif TEST2
#define	setpoint(x,y) curbuf.block[pointbyte(x,y)]|=pointbit(x,y)
#define SOLID -1
#define DOTTED 014
#define SHORTDASHED 034
#define DOTDASHED 054
#define LONGDASHED 074

extern char *mktemp();

int	again;		/* is there more plotting to do? */
int	done1;		/* any plotting done? */

/* plot state variables */
int	linmod	= SOLID;
int	lastx;
int	lasty;
int	radius, startx, starty, endx, endy;

/* logical plotting area - assumed to be square */
double	botx, topx;
double	boty, topy;
double	del;	/* length & width of plot in logical plot units */
/* raster offsets for centering of plotting area on device */
int	centx = 0;
int	centy = 0;

int warned = 0;	/* Indicates whether the warning message about
		 * unimplemented routines has been printed */

typedef struct	buf
{
	int	bno;
	char	*block;
};
struct	buf	curbuf = { -1, (char *)0 };
char	blocks	[NB][BSIZ];
#ifndef	VAX
struct	buf	bufs[NB];
#endif

int	in, out;
char	*picture = "/usr/tmp/#rastXXXXX";

/* variables for used to print from font file */
int fontSet = 0;	/* Has the font file been read */
struct header header;
struct dispatch dispatch[256];
char *bits;
char *fontFile = "/usr/lib/vfont1/R.6";

FILE *infile;
#ifndef QUE_RASTER
int	vfd;		/* versatec output file descriptor */
#endif QUE_RASTER

main( argc, argv )
int argc;
char **argv;
{
	extern int onintr();
	register int	b;

	infile = stdin;
#ifndef	DEBUG
#ifndef QUE_RASTER
	vfd = fileno( stdout );
#endif QUE_RASTER
#endif	DEBUG
	for ( b=1; argv[b][0] == '-'; b++ )
	{
		switch ( argv[b][1] )
		{
		case 'o':
			putpict( argv[++b] );
			if ( argc <= b )
			{
			    fprintf( stderr, "usage: %s -o rast_file\n",
					argv[0] );
			    exit(1);
			}
			break;
		default:
			fprintf( stderr, "%s: %s option unknown\n",
				argv[0], argv[b] );
			break;
		}
	}

	signal(SIGTERM, onintr);
	if (signal(SIGINT, SIG_IGN) != SIG_IGN)
		signal(SIGINT, onintr);
	strcpy( picture, mktemp( "/usr/tmp/#rastXXXXX" ) );
	while ( argc>b )
	{
		plotfile( argv[b++] );
	}
	exit(0);
}

/* plot named file ---
 *
 */
plotfile( filename )
char	*filename;		/* name of plot file */
{
	register int	i;

#ifdef	DEBUG
	fprintf( stderr, "plot file: %s\n", filename );
#endif
	if ( ( infile = fopen( filename, "r" )) == NULL )
	{
		perror(filename);
		exit(1);
	}
	done1 = 0;	/* no plots from this file yet */
	/* init constants for scaling */
	topx = topy = (double)RASTERSIZE;
	botx = boty = (double)0;
	del = (double)RASTERSIZE;
	centx = centy = 0;
	centx = (int)((RASTERSIZE - mapx(topx))/2);
	centy = (int)(mapy(topy)/2);
#ifdef	DEBUG
	print_sizes();
#endif
	do
	{
#ifdef		DEBUG
		fprintf( stderr, "new picture\n" );
#endif
		out = creat(picture, 0600);
		in = open(picture, 0);

#ifndef		VAX
		for (i=0; i<NB; i++)
		{
			bufs[i].bno = -1;
			bufs[i].block = blocks[i];
		}
#endif
#ifdef		HOLE_PROBLEM
/*delete following code when filsys deals properly with holes in files*/
		for(i=0;i<BSIZ;i++)
			blocks[0][i] = 0;
		zseek(out, 0);
		for(i=0;i<32*32;i++)
			write(out,blocks[0],BSIZ);
/**/
#endif
		curbuf.bno = -1;
		curbuf.block = (char *)0;
		if ( getpict() == 1 ) /* error */
		{
			quit();
		}

#ifdef		VAX
		for (i=0; i<NB; i++)
		{
			zseek(out,i);
			write(out, blocks[i], BSIZ);
		}
#else
		for (i=0; i<NB; i++)
			if (bufs[i].bno != -1)
			{
				zseek(out, bufs[i].bno);
				write(out, bufs[i].block, BSIZ);
			}
#endif
		close(in);
		close(out);
		putpict(picture);
		unlink(picture);
		clean_blocks();
	} while ( again );
	return;
}

quit()
{
	close(in);
	close(out);
	unlink(picture);
	exit(1);
}
clean_blocks()
{
	register int	b, i;

	for ( b=0; b<NB; b++ )
	{
		for ( i=0; i<BSIZ; i++ )
			blocks[b][i]='\0';
	}
}

getpict()
{
	register x1, y1;
	double	delx;
	double	dely;

	again = 0;
	for (;;)
	switch (x1 = getc(infile))
	{
	case '\n':
		break;

	case 's':
		botx = getinteger(infile);
		boty = getinteger(infile);
		topx = getinteger(infile);
		topy = getinteger(infile);
		delx = topx-botx;
		dely = topy-boty;
		if ( dely/delx > 1 )
			del = dely;
		else
			del = delx;
		centx = 0;
		centx = (RASTERSIZE - mapx(topx))/2;
		centy = 0;
		centy = mapy(topy) / 2;
#ifdef	DEBUG
	print_sizes();
#endif
		continue;

	case 'b':
		x1 = getc(infile) ;
		break ;
	case 'l':
		done1 |= 01;
		x1 = mapx(getinteger(infile));
		y1 = mapy(getinteger(infile));
		lastx = mapx(getinteger(infile));
		lasty = mapy(getinteger(infile));
		line(x1, y1, lastx, lasty);
		continue;

	case 'c':
		x1 = mapx(getinteger(infile));
		y1 = mapy(getinteger(infile));
		radius = mapx(getinteger(infile));
		if(!warned)
		{
			fprintf(stderr,"Circles are Implemented\n");
			warned++;
		}
		circle(x1, y1, radius);
		continue;
		
	case 'a':
		x1 = mapx(getinteger(infile));
		y1 = mapy(getinteger(infile));
		startx = mapx(getinteger(infile));
		starty = mapy(getinteger(infile));
		endx = mapx(getinteger(infile));
		endy = mapy(getinteger(infile));
		if(!warned)
		{
			fprintf(stderr,"Circles and Arcs are unimplemented\n");
			warned++;
		}
		continue;

	case 'm':
		lastx = mapx(getinteger(infile));
		lasty = mapy(getinteger(infile));
		continue;

	case 't':
		lastx = lastx - 6;
		lasty = lasty + 6;
		done1 |= 01;
		while ((x1 = getc(infile)) != '\n')
			plotch(x1);
		continue;

	case 'e':
		if (done1)
		{
			again++;
			return(0);
		}
		continue;

	case 'p':
		done1 |= 01;
		lastx = mapx(getinteger(infile));
		lasty = mapy(getinteger(infile));
		point(lastx, lasty);
		point(lastx+1, lasty);
		point(lastx, lasty+1);
		point(lastx+1, lasty+1);
		continue;

	case 'n':
		done1 |= 01;
		x1 = mapx(getinteger(infile));
		y1 = mapy(getinteger(infile));
		line(lastx, lasty, x1, y1);
		lastx = x1;
		lasty = y1;
		continue;

	case 'f':
		getinteger(infile);
		getc(infile);
		switch(getc(infile))
		{
		case 't':
			linmod = DOTTED;
			break;
		default:
		case 'i':
			linmod = SOLID;
			break;
		case 'g':
			linmod = LONGDASHED;
			break;
		case 'r':
			linmod = SHORTDASHED;
			break;
		case 'd':
			linmod = DOTDASHED;
			break;
		}
		while((x1=getc(infile))!='\n')
			if(x1==-1) return(0);
		continue;

	case 'd':
		getinteger(infile);
		getinteger(infile);
		getinteger(infile);
		x1 = getinteger(infile);
		while (--x1 >= 0)
			getinteger(infile);
		continue;

	case 0:		/* ignore null characters */
		continue;

	case 255:
	case -1:
		return(0);

	default:
		printf("Botch %c(%o)\n",x1,x1);
		return(1);
	}
}

plotch(ch)
char ch;
{
	register int i,j,k;
	register char *ptr,c;
	int nbytes;

	if(!fontSet) InitFont();	/* Read font if not already read */

	ptr = bits + dispatch[ch].addr;

	for(i = dispatch[ch].up; i > -dispatch[ch].down; --i)
	{
		nbytes = (dispatch[ch].right + dispatch[ch].left + 7)/8;
		for(j = 0; j < nbytes; j++)
		{
			c = *ptr++;
			for(k=7; k>=0; --k)
				if((c>>k) & 1)
					point(lastx+7-k+j*8-dispatch[ch].left,
						lasty-i);
		}
	}
	if(ch != ' ') lastx += dispatch[ch].width;
	else lastx +=dispatch['a'].width;
}

InitFont()
{
	char *s;
	int fonts;
	int i;

	fontSet = 1;
	/* Get the font file */
	s = fontFile;
	if((fonts = open(s,0)) == -1)
	{
		perror(s);
		fprintf(stderr,"Can't get font file");
		abort();
	}
	/* Get the header and check magic number */
	if(read(fonts,&header,sizeof(header)) != sizeof(header))
	{
		perror(s);
		fprintf(stderr,"Bad read in font file");
		abort();
	}
	if(header.magic != 0436)
	{
		fprintf(stderr,"Bad magic numer in font file");
		abort();
	}
	/* Get dispatches */
	if(read(fonts,dispatch,sizeof(dispatch)) != sizeof(dispatch))
	{
		perror(s);
		fprintf(stderr,"Bad read in font file");
		abort();
	}
	/* Allocate space for bit map and read in bits */
	bits = (char *) malloc(header.size);
	if(read(fonts,bits,header.size) != header.size)
	{
		perror(s);
		fprintf(stderr,"Can't read bit map in font file");
		abort();
	}
	/* Close font file */
	if(close(fonts) != 0)
	{
		perror(s);
		fprintf(stderr,"Can't close font file");
		abort();
	}
}

#ifndef QUE_RASTER
/* globals from vpltdmp.c */
int	plotmd[]	= { VPLOT, 0, 0};
int	prtmd[]		= { VPRINT, 0, 0};
#ifndef TEST2
int	obuf[220*64];		/* output buffer; 64 lines of output */
#else TEST2
char	obuf[IO_BSIZ];	/* output buffer */
#endif TEST2

putpict(picture)
char *picture;
{
	FILE *raster;
	int	in;
#ifndef TEST2
	register x, *ip, *op;
	int i,y;
#else TEST2
	int	nbytes;

#endif TEST2

#ifdef	DEBUG
	return;		/* don't run raster */
#endif	DEBUG
	if((raster = fopen(picture,"r")) == NULL)
	{
		perror(picture);
		exit(1);
	}

	ioctl( vfd, VSETSTATE, plotmd);
	in = fileno( raster );
	lseek( in, 0L, 0);
#ifndef TEST2
	for (i=0; i<32; i++)
	{
		read( in, blocks[0], 32*BSIZ);
		for (y=0; y<64; y++)
		{
			op = &(obuf[(y*BYTES_PER_LINE)/4]);
			for (x=0; x<32; x++) 
			{
				ip = (int *)&blocks[x][y<<3];
				*op++ = *ip++;
				*op++ = *ip++;
			}
		}
		write( vfd, (char *)obuf, BYTES_PER_LINE*64);
	}
#else TEST2
	while ( ( nbytes=read( in, obuf, IO_BSIZ )) > 0 )
	{
		write( vfd, obuf, nbytes );
	}
#endif TEST2
	fclose(raster);
	ioctl(vfd, VSETSTATE,prtmd);
}
#else	QUE_RASTER
putpict(picture)
char *picture;
{
	char cmdbuf[BSIZ];

	sprintf( cmdbuf, "qpr -q vplot -f 132col %s", picture );
	system( cmdbuf );
}
#endif	QUE_RASTER

line(x0, y0, x1, y1)
register x0, y0;
{
	int dx, dy;
	int xinc, yinc;
	register res1;
	int res2;
	int slope;

	xinc = 1;
	yinc = 1;
	if ((dx = x1-x0) < 0) {
		xinc = -1;
		dx = -dx;
	}
	if ((dy = y1-y0) < 0) {
		yinc = -1;
		dy = -dy;
	}
	slope = xinc*yinc;
	res1 = 0;
	res2 = 0;
	if (dx >= dy) while (x0 != x1) {
	if((x0+slope*y0)&linmod)
	if (blockno(x0,y0) == curbuf.bno)
		setpoint( x0, y0 );
	else
		point(x0, y0);
		if (res1 > res2) {
			res2 += dx - res1;
			res1 = 0;
			y0 += yinc;
		}
		res1 += dy;
		x0 += xinc;
	} else while (y0 != y1) {
	if((x0+slope*y0)&linmod)
	if (blockno(x0,y0) == curbuf.bno)
		setpoint( x0, y0 );
	else
		point(x0, y0);
		if (res1 > res2) {
			res2 += dy - res1;
			res1 = 0;
			x0 += xinc;
		}
		res1 += dx;
		y0 += yinc;
	}
	if((x1+slope*y1)&linmod)
	if (blockno(x1,y1) == curbuf.bno)
		setpoint( x1, y1 );
	else
		point(x1, y1);
}

point(x, y)
register x, y;
{
	register bno;

#ifdef	POINT_DEBUG
	fprintf( stderr, "(x,y)=(%d,%d) block=%d byte=%d bit=%o\n",
		x, y, blockno(x,y), pointbyte(x,y), pointbit(x,y) );
#endif	POINT_DEBUG
	bno = blockno(x,y);
	if (bno != curbuf.bno) {
		if (bno < 0 || bno >= TOTALBLOCKS)
			return;
		getblk(bno);
	}
	setpoint( x, y );
}

#ifndef VAX
getblk(b)
register b;
{
	register struct buf *bp1, *bp2;
	register char *tp;

	if(b < 0 || b >= TOTALBLOCKS)
	{
		fprintf(stderr,
			"vplot: internal error, b out of range in getblk\n");
		abort();
	}
loop:
	for (bp1 = bufs; bp1 < &bufs[NB]; bp1++) {
		if (bp1->bno == b || bp1->bno == -1) {
			tp = bp1->block;
			for (bp2 = bp1; bp2>bufs; --bp2) {
				bp2->bno = (bp2-1)->bno;
				bp2->block = (bp2-1)->block;
			}
			curbuf.bno = b;
			curbuf.block = tp;
			return;
		}
	}
	zseek(out, bufs[NB-1].bno);
	write(out, bufs[NB-1].block, BSIZ);
	zseek(in, b);
	read(in, bufs[NB-1].block, BSIZ);
	bufs[NB-1].bno = b;
	goto loop;
}
#else	VAX
getblk(b)
register b;
{
	if(b < 0 || b >= TOTALBLOCKS)
	{
		fprintf(stderr,
			"vplot: internal error, b out of range in getblk\n");
		abort();
	}
	curbuf.bno = b;
	curbuf.block = blocks[b];
}
#endif	VAX

onintr()
{
	exit(1);
}

zseek(a, b)
{
	return(lseek(a, (long)b*512, 0));
}


getinteger(f)
FILE *f;
{
	register int low,high,result;
	low = getc(f);
	high = getc(f);
	result = ((high*256) | low);
	if(high>127) result = 0xffff0000 | result;
	return(result);
}
circle(x,y,c)
{
	register dx, dy;
	long ep;
	long labs();
	int de;
	dx = 0;
	ep = 0;
	for(dy=c; dy>=dx; dy--)
	{
		for(;;)
		{
			point(x+dx, y+dy);
			point(x-dx, y+dy);
			point(x+dx, y-dy);
			point(x-dx, y-dy);
			point(x+dy, y+dx);
			point(x-dy, y+dx);
			point(x+dy, y-dx);
			point(x-dy, y-dx);
			ep += 2*dx + 1;
			de = -2*dy + 1;
			dx++;
			if(labs(ep)>=labs(ep+de))
			{
				ep += de;
				break;
			}
		}
	}
}

long labs(a)
long a;
{
	return(a>=0? a: -a);
}

/* print plotting dimension info to stderr, for debug */
print_sizes()
{
	fprintf( stderr, "raster:%d\n", RASTERSIZE );
	fprintf( stderr, "x:%f-%f\ty:%f-%f\n", botx, topx, boty, topy );
	fprintf( stderr, "del:%f\n", del );
	fprintf( stderr, "centx:%d\tcenty:%d\n", centx, centy );
}
